@@ -1,5 +1,3 @@
-from __future__ import absolute_import, print_function
-
 import distutils.version
 import os
 import re
@@ -200,7 +198,7 @@ def has_rhg():
 
 
 @check("pyoxidizer", "running with pyoxidizer build as 'hg'")
-def has_rhg():
+def has_pyoxidizer():
     return 'PYOXIDIZED_INSTALLED_AS_HG' in os.environ
 
 
@@ -410,7 +408,7 @@ def getgitversion():
 
 
 @check("pygit2", "pygit2 Python library")
-def has_git():
+def has_pygit2():
     try:
         import pygit2
 
@@ -752,7 +750,7 @@ def has_test_repo():
 
 
 @check("network-io", "whether tests are allowed to access 3rd party services")
-def has_test_repo():
+def has_network_io():
     t = os.environ.get("HGTESTS_ALLOW_NETIO")
     return t == "1"
 
@@ -1123,7 +1121,7 @@ def has_emacs():
 @check('black', 'the black formatter for python (>= 20.8b1)')
 def has_black():
     blackcmd = 'black --version'
-    version_regex = b'black, version ([0-9a-b.]+)'
+    version_regex = b'black, (?:version )?([0-9a-b.]+)'
     version = matchoutput(blackcmd, version_regex)
     sv = distutils.version.StrictVersion
     return version and sv(_bytes2sys(version.group(1))) >= sv('20.8b1')
